"""Contact client for registry service."""
from datetime import datetime
from enum import Enum, IntEnum, unique
from typing import Any, Dict, List, Optional, Union

import fred_api.registry.contact.service_contact_grpc_pb2_grpc
import grpc
from fred_api.registry.contact.contact_history_types_pb2 import ContactHistoryReply, ContactHistoryRequest
from fred_api.registry.contact.contact_info_types_pb2 import (
    ContactAdditionalIdentifier as AdditionalIdentifierMessage, ContactIdReply, ContactIdRequest, ContactInfoReply,
    ContactInfoRequest, WarningLetter as WarningLetterMessage)
from fred_api.registry.contact.contact_state_history_types_pb2 import (ContactStateHistoryReply,
                                                                       ContactStateHistoryRequest)
from fred_api.registry.contact.contact_state_types_pb2 import ContactStateReply, ContactStateRequest
from pydantic import BaseModel

from .common import Address, ObjectEvents, RegistryDecoder
from .exceptions import ContactDoesNotExist, InvalidHistoryInterval
from .object import BaseObjectHistoryId, BaseObjectId, RegistryObjectClient


class ContactId(BaseObjectId):
    """Contact id."""


class ContactHistoryId(BaseObjectHistoryId):
    """Contact history id."""


@unique
class WarningLetterPreference(IntEnum):
    """Enum with warning letter results."""

    TO_SEND = 0
    NOT_TO_SEND = 1
    NOT_SPECIFIED = 2


@unique
class ContactIdentifierType(str, Enum):
    """Contact identifier type enum."""

    NATIONAL_IDENTITY_NUMBER = 'national_identity_number'
    NATIONAL_IDENTITY_CARD = 'national_identity_card'
    PASSPORT_NUMBER = 'passport_number'
    COMPANY_REGISTRATION_NUMBER = 'company_registration_number'
    SOCIAL_SECURITY_NUMBER = 'social_security_number'
    BIRTHDATE = 'birthdate'


class AdditionalIdentifier(BaseModel):
    """Additional contact identifier."""

    type: ContactIdentifierType
    value: str


class Contact(BaseModel):
    """Contact info."""

    contact_id: ContactId
    contact_handle: str
    contact_history_id: Optional[ContactHistoryId] = None
    name: str = ''
    organization: str = ''
    place: Optional[Address] = None
    telephone: Optional[str] = None
    fax: Optional[str] = None
    emails: List[str] = []
    notify_emails: List[str] = []
    vat_identification_number: Optional[str] = None
    additional_identifier: Optional[AdditionalIdentifier] = None
    publish: Dict[str, bool] = {}
    mailing_address: Optional[Address] = None
    billing_address: Optional[Address] = None
    shipping_address1: Optional[Address] = None
    shipping_address2: Optional[Address] = None
    shipping_address3: Optional[Address] = None
    warning_letter: WarningLetterPreference = WarningLetterPreference.NOT_SPECIFIED
    auth_info: Optional[str] = None
    events: Optional[ObjectEvents] = None


class ContactDecoder(RegistryDecoder):
    """Contact service decoder."""

    def __init__(self) -> None:
        super().__init__()
        self.set_decoder(AdditionalIdentifierMessage, self._decode_additional_identifier)
        self.set_decoder(WarningLetterMessage, self._decode_warning_letter)
        self.set_exception_decoder(ContactIdReply.Exception.ContactDoesNotExist, ContactDoesNotExist)
        self.set_exception_decoder(ContactInfoReply.Exception.ContactDoesNotExist, ContactDoesNotExist)
        self.set_exception_decoder(ContactHistoryReply.Exception.ContactDoesNotExist, ContactDoesNotExist)
        self.set_exception_decoder(ContactHistoryReply.Exception.InvalidHistoryInterval, InvalidHistoryInterval)
        self.set_exception_decoder(ContactStateReply.Exception.ContactDoesNotExist, ContactDoesNotExist)
        self.set_exception_decoder(ContactStateHistoryReply.Exception.ContactDoesNotExist, ContactDoesNotExist)
        self.set_exception_decoder(ContactStateHistoryReply.Exception.InvalidHistoryInterval, InvalidHistoryInterval)

    def _decode_additional_identifier(self, value: AdditionalIdentifierMessage) -> Optional[AdditionalIdentifier]:
        """Decode additional identifier."""
        field = value.WhichOneof('Variant')
        if field is not None:
            return AdditionalIdentifier(type=ContactIdentifierType(field), value=self.decode(getattr(value, field)))
        else:
            return None

    def _decode_warning_letter(self, value: WarningLetterMessage) -> WarningLetterPreference:
        """Decode WarningLetter."""
        return WarningLetterPreference(value.preference)


class ContactClient(RegistryObjectClient[ContactId, ContactHistoryId]):
    """Contact service gRPC client."""

    service = 'Contact'
    object_id = 'contact_id'
    object_history_id = 'contact_history_id'
    object_handle = 'contact_handle'

    def __init__(self, netloc: str, credentials: grpc.ChannelCredentials = None) -> None:
        """Initialize contact client instance.

        Args:
            netloc: Netloc of contact service server.
            credentials: Optional credentials.
        """
        super().__init__(
            netloc,
            decoder=ContactDecoder(),
            grpc_modules=[fred_api.registry.contact.service_contact_grpc_pb2_grpc],
            credentials=credentials,
        )

    async def get_contact_id(self, handle: str) -> ContactId:
        """Get contact id.

        Args:
            handle: Contact handle.

        Raises:
            ContactDoesNotExist: If contact does not exist.
        """
        return ContactId(await self.get_object_id(request=ContactIdRequest(), method='get_contact_id', handle=handle))

    async def get_contact_info(self, contact_id: ContactId, history_id: ContactHistoryId = None) -> Contact:
        """Get contact info.

        Args:
            contact_id: Contact ID.
            history_id: Contact history ID.

        Raises:
            ContactDoesNotExist: If contact does not exist.
        """
        result = await self.get_object_info(
            request=ContactInfoRequest(), method='get_contact_info',
            object_id=contact_id, history_id=history_id,
        )
        return Contact(**result)

    async def get_contact_state(self, contact_id: ContactId) -> Dict[str, bool]:
        """Get contact state consisting of state flags and their boolean values.

        Raises:
            ContactDoesNotExist: If contact does not exist.

        Args:
            contact_id: Contact ID.
        """
        return await self.get_object_state(
            request=ContactStateRequest(),
            method='get_contact_state',
            object_id=contact_id,
        )

    async def get_contact_history(
        self, contact_id: ContactId, start: Union[ContactHistoryId, datetime] = None,
        end: Union[ContactHistoryId, datetime] = None,
    ) -> Dict[str, Any]:
        """Get contact history timeline.

        Args:
            contact_id: Contact ID.
            start: Datetime or history_id of history interval start.
            end: Datetime or history_id of history interval end.

        Raises:
            ContactDoesNotExist: If contact does not exist.
            InvalidHistoryInterval: In case of invalid history interval.
        """
        return await self.get_object_history(
            request=ContactHistoryRequest(), method='get_contact_history',
            object_id=contact_id, start=start, end=end,
        )

    async def get_contact_state_history(
        self, contact_id: ContactId, start: Union[ContactHistoryId, datetime] = None,
        end: Union[ContactHistoryId, datetime] = None,
    ) -> Dict[str, Any]:
        """Get contact state history timeline.

        Args:
            contact_id: Contact ID.
            start: Datetime or history_id of history interval start.
            end: Datetime or history_id of history interval end.

        Raises:
            ContactDoesNotExist: If contact does not exist.
            InvalidHistoryInterval: In case of invalid history interval.
        """
        return await self.get_object_state_history(
            request=ContactStateHistoryRequest(), method='get_contact_state_history',
            object_id=contact_id, start=start, end=end,
        )
