"""Domain client for registry Domain service."""
from datetime import date, datetime
from enum import IntEnum, unique
from typing import Any, AsyncIterable, Dict, Optional, Sequence, Union, cast

import fred_api.registry.domain.service_admin_grpc_pb2_grpc
import fred_api.registry.domain.service_domain_grpc_pb2_grpc
import grpc
from aioitertools.more_itertools import chunked
from aioitertools.types import AnyIterable
from fred_api.registry.domain.domain_history_types_pb2 import DomainHistoryReply, DomainHistoryRequest
from fred_api.registry.domain.domain_info_types_pb2 import (DomainIdReply, DomainIdRequest, DomainInfoReply,
                                                            DomainInfoRequest)
from fred_api.registry.domain.domain_state_history_types_pb2 import DomainStateHistoryReply, DomainStateHistoryRequest
from fred_api.registry.domain.domain_state_types_pb2 import DomainStateReply, DomainStateRequest
from fred_api.registry.domain.service_admin_grpc_pb2 import (BatchDeleteDomainsReply, BatchDeleteDomainsRequest,
                                                             ManageDomainStateFlagsReply, ManageDomainStateFlagsRequest)
from frgal.aio import AsyncGrpcClient
from pydantic import BaseModel

from .common import DatetimeRange, ObjectEvents, RegistryDecoder
from .contact import ContactId
from .exceptions import DomainDoesNotExist, InvalidHistoryInterval
from .keyset import KeysetId
from .nsset import NssetId
from .object import BaseObjectHistoryId, BaseObjectId, RegistryObjectClient


class DomainId(BaseObjectId):
    """Domain id."""


class DomainHistoryId(BaseObjectHistoryId):
    """Domain history id."""


class Domain(BaseModel):
    """Domain info."""

    domain_id: DomainId
    fqdn: str
    domain_history_id: Optional[DomainHistoryId] = None
    registrant: Optional[ContactId] = None
    sponsoring_registrar: Optional[str] = None
    nsset: Optional[NssetId] = None
    keyset: Optional[KeysetId] = None
    administrative_contacts: Sequence[ContactId] = []
    expires_at: Optional[date] = None
    auth_info: Optional[str] = None
    events: Optional[ObjectEvents] = None


@unique
class NotDeletedReason(IntEnum):
    """Enum with reasons for domain not to be deleted."""

    UNDEFINED = 0
    DOMAIN_BLOCKED = 1
    DOMAIN_DOES_NOT_EXIST = 2
    DOMAIN_HISTORY_ID_MISMATCH = 3
    INSUFFICIENT_PERMISSIONS = 4


class DomainDecoder(RegistryDecoder):
    """Domain service decoder."""

    def __init__(self) -> None:
        super().__init__()
        self.set_exception_decoder(DomainIdReply.Exception.DomainDoesNotExist, DomainDoesNotExist)
        self.set_exception_decoder(DomainInfoReply.Exception.DomainDoesNotExist, DomainDoesNotExist)
        self.set_exception_decoder(DomainHistoryReply.Exception.DomainDoesNotExist, DomainDoesNotExist)
        self.set_exception_decoder(DomainHistoryReply.Exception.InvalidHistoryInterval, InvalidHistoryInterval)
        self.set_exception_decoder(DomainStateReply.Exception.DomainDoesNotExist, DomainDoesNotExist)
        self.set_exception_decoder(DomainStateHistoryReply.Exception.DomainDoesNotExist, DomainDoesNotExist)
        self.set_exception_decoder(DomainStateHistoryReply.Exception.InvalidHistoryInterval, InvalidHistoryInterval)


class DomainClient(RegistryObjectClient[DomainId, DomainHistoryId]):
    """Domain service gRPC client."""

    service = 'Domain'
    object_id = 'domain_id'
    object_history_id = 'domain_history_id'
    object_handle = 'fqdn'

    def __init__(self, netloc: str, credentials: grpc.ChannelCredentials = None) -> None:
        """Initialize domain client instance.

        Args:
            netloc: Netloc of Domain service server.
            credentials: Optional credentials.
        """
        super().__init__(
            netloc,
            decoder=DomainDecoder(),
            grpc_modules=[fred_api.registry.domain.service_domain_grpc_pb2_grpc],
            credentials=credentials,
        )

    async def get_domain_id(self, fqdn: str) -> DomainId:
        """Get domain id.

        Args:
            fqdn: Domain FQDN.

        Raises:
            DomainDoesNotExist: If domain does not exist.
        """
        return DomainId(await self.get_object_id(request=DomainIdRequest(), method='get_domain_id', handle=fqdn))

    async def get_domain_info(self, domain_id: DomainId, history_id: DomainHistoryId = None) -> Domain:
        """Get domain info.

        Args:
            domain_id: Domain ID.
            history_id: Domain history ID.

        Raises:
            DomainDoesNotExist: If domain does not exist.
        """
        result = await self.get_object_info(
            request=DomainInfoRequest(), method='get_domain_info',
            object_id=domain_id, history_id=history_id,
        )
        expires_at = result.pop('expires_at')
        if expires_at:
            expires_at = expires_at.date()  # Convert expiry to date
        return Domain(expires_at=expires_at, **result)

    async def get_domain_state(self, domain_id: DomainId) -> Dict[str, bool]:
        """Get domain state consisting of state flags and their boolean values.

        Raises:
            DomainDoesNotExist: If domain does not exist.

        Args:
            domain_id: Domain ID.
        """
        return await self.get_object_state(request=DomainStateRequest(), method='get_domain_state', object_id=domain_id)

    async def get_domain_history(
        self, domain_id: DomainId, start: Union[DomainHistoryId, datetime] = None,
        end: Union[DomainHistoryId, datetime] = None,
    ) -> Dict[str, Any]:
        """Get domain history timeline.

        Args:
            domain_id: Domain ID.
            start: Datetime or history_id of history interval start.
            end: Datetime or history_id of history interval end.

        Raises:
            DomainDoesNotExist: If domain does not exist.
            InvalidHistoryInterval: In case of invalid history interval.
        """
        return await self.get_object_history(
            request=DomainHistoryRequest(), method='get_domain_history',
            object_id=domain_id, start=start, end=end,
        )

    async def get_domain_state_history(
        self, domain_id: DomainId, start: Union[DomainHistoryId, datetime] = None,
        end: Union[DomainHistoryId, datetime] = None,
    ) -> Dict[str, Any]:
        """Get domain state history timeline.

        Args:
            domain_id: Domain ID.
            start: Datetime or history_id of history interval start.
            end: Datetime or history_id of history interval end.

        Raises:
            DomainDoesNotExist: If domain does not exist.
            InvalidHistoryInterval: In case of invalid history interval.
        """
        return await self.get_object_state_history(
            request=DomainStateHistoryRequest(), method='get_domain_state_history',
            object_id=domain_id, start=start, end=end,
        )


class DomainAdminDecoder(RegistryDecoder):
    """Domain Admin service decoder."""

    def __init__(self) -> None:
        super().__init__()
        self.set_decoder(BatchDeleteDomainsReply.NotDeleted, self._decode_not_deleted)

        self.set_exception_decoder(ManageDomainStateFlagsReply.Exception.DomainDoesNotExist, DomainDoesNotExist)
        self.set_exception_decoder(ManageDomainStateFlagsReply.Exception.DomainStateFlagRequestDoesNotExist, ValueError)
        self.set_exception_decoder(ManageDomainStateFlagsReply.Exception.DomainStateFlagDoesNotExist, ValueError)

    def _decode_not_deleted(self, value: BatchDeleteDomainsReply.NotDeleted) -> dict:
        decoded = cast(Dict[str, Any], self._decode_message(value))
        decoded['reason'] = NotDeletedReason(decoded['reason'])
        return decoded


class DomainAdminClient(AsyncGrpcClient):
    """Domain Admin service gRPC client."""

    service = 'Admin'
    decoder_cls = DomainAdminDecoder
    grpc_modules = (fred_api.registry.domain.service_admin_grpc_pb2_grpc,)

    @staticmethod
    async def _make_batch_delete_domains_requests(
        domains: AnyIterable[Domain], *, batch_size: int = 100
    ) -> AsyncIterable[BatchDeleteDomainsRequest]:
        """Request generator for `batch_delete_domains` method."""
        async for chunk in chunked(domains, batch_size):
            request = BatchDeleteDomainsRequest()
            for domain in chunk:
                message = BatchDeleteDomainsRequest.Domain()
                message.domain_id.uuid.value = domain.domain_id
                if domain.domain_history_id is None:
                    raise ValueError("{!r} is missing a domain_history_id")
                message.domain_history_id.uuid.value = domain.domain_history_id
                request.domains.append(message)
            yield request

    async def batch_delete_domains(self, domains: AnyIterable[Domain], *, batch_size: int = 100) -> AsyncIterable:
        """Delete batch of domains.

        Args:
            domains: Iterable (or async iterable) of domains to be deleted.
            batch_size: Number of domains in single stream request.
        """
        requests = self._make_batch_delete_domains_requests(domains, batch_size=batch_size)
        async for message in self.call_stream(self.service, 'batch_delete_domains', requests):  # pragma: no cover
            for domain in message or ():
                yield domain

    @staticmethod
    async def _make_manage_domain_state_flags_requests(
        domains: AnyIterable[Domain],
        state_flags: Dict[str, Optional[DatetimeRange]],
        close_requests: AnyIterable[str],
        *,
        batch_size: int = 100
    ) -> AsyncIterable[ManageDomainStateFlagsRequest]:
        """Request generator for `manage_domain_state_flags` method."""
        async for chunk in chunked(domains, batch_size):
            request = ManageDomainStateFlagsRequest()
            for flag, validity in state_flags.items():
                validity_msg = request.state_flag_requests_to_create.state_flags.get_or_create(flag)
                if validity and validity.start:
                    validity_msg.valid_from.FromDatetime(validity.start)
                if validity and validity.end:
                    validity_msg.valid_to.FromDatetime(validity.end)
            for domain in chunk:
                message = ManageDomainStateFlagsRequest.Domain()
                message.domain_id.uuid.value = domain.domain_id
                if domain.domain_history_id:
                    message.domain_history_id.uuid.value = domain.domain_history_id
                request.state_flag_requests_to_create.domains.append(message)
            yield request

        async for close_chunk in chunked(close_requests, batch_size):  # pragma: no cover # because of py39 coverage
            request = ManageDomainStateFlagsRequest()
            for request_id in close_chunk:
                message = request.state_flag_requests_to_close.add()
                message.uuid.value = request_id
            yield request

    async def manage_domain_state_flags(
        self,
        *,
        domains: AnyIterable[Domain] = None,
        state_flags: Dict[str, Optional[DatetimeRange]] = None,
        close_requests: AnyIterable[str] = None,
        batch_size: int = 100
    ) -> AsyncIterable[Dict[str, str]]:
        """Manage domains state flags.

        Args:
            domains: Iterable (or async iterable) of domains to be changed.
            state_flags: Dictionary. Keys are state flags, values are datetime ranges
                (or None if datetime range is from NOW to indefinitely).
            close_requests: String ids of domain state flag requests that should be closed.
            batch_size: Number of domains in single stream request.

        Returns:
            Async iterable of dictionaries with following keys:
                request_id: Contains ID of flag state request. It can be used to close the request later.
                domain_id: Domain ID.
                state_flag: Name of state flag that was set.

        Raises:
            DomainDoesNotExist: When domain does not exist.
            ValueError: When state flag or state flag request id does not exist.
        """
        if bool(domains) ^ bool(state_flags):
            raise ValueError('Either both `domains` and `state_flags` have to be defined or none of them.')

        requests = self._make_manage_domain_state_flags_requests(
            domains or [], state_flags or {}, close_requests or [], batch_size=batch_size,
        )
        async for message in self.call_stream(self.service, 'manage_domain_state_flags', requests):  # pragma: no cover
            for record in message or []:
                record['domain_id'] = DomainId(record['domain_id'])
                yield record
