"""Registrar client for registry service."""
from decimal import Decimal
from typing import Dict, List, Optional, Sequence, cast

import fred_api.registry.registrar.service_registrar_grpc_pb2_grpc
import grpc
from fred_api.registry.registrar.registrar_common_types_pb2 import Credit
from fred_api.registry.registrar.registrar_credit_types_pb2 import RegistrarCreditReply, RegistrarCreditRequest
from fred_api.registry.registrar.registrar_info_types_pb2 import RegistrarInfoReply, RegistrarInfoRequest
from fred_api.registry.registrar.registrar_zone_access_types_pb2 import (RegistrarZoneAccessReply,
                                                                         RegistrarZoneAccessRequest)
from frgal.aio import AsyncGrpcClient
from google.protobuf.empty_pb2 import Empty
from pydantic import BaseModel

from .common import Address, RegistryDecoder
from .exceptions import RegistrarDoesNotExist


class Registrar(BaseModel):
    """Registrar info."""

    registrar_handle: str
    name: str
    organization: str = ''
    place: Optional[Address] = None

    telephone: Optional[str] = None
    fax: Optional[str] = None
    emails: Sequence[str] = []
    url: Optional[str] = None
    company_registration_number: Optional[str] = None
    vat_identification_number: Optional[str] = None

    is_system_registrar: bool = False

    variable_symbol: str = ''
    payment_memo_regex: str = ''
    is_vat_payer: bool = False


class RegistrarDecoder(RegistryDecoder):
    """Registrar service decoder."""

    def __init__(self) -> None:
        super().__init__()
        self.set_decoder(Credit, self._decode_credit)
        self.set_exception_decoder(RegistrarInfoReply.Exception.RegistrarDoesNotExist, RegistrarDoesNotExist)
        self.set_exception_decoder(RegistrarCreditReply.Exception.RegistrarDoesNotExist, RegistrarDoesNotExist)
        self.set_exception_decoder(RegistrarZoneAccessReply.Exception.RegistrarDoesNotExist, RegistrarDoesNotExist)

    def _decode_credit(self, value: Credit) -> Optional[Decimal]:
        """Decode registrar credit."""
        if not value.decimal_value:
            return None
        return Decimal(value.decimal_value)


class RegistrarClient(AsyncGrpcClient):
    """Registrar service gRPC client."""

    service = 'Registrar'

    def __init__(self, netloc: str, credentials: grpc.ChannelCredentials = None) -> None:
        """Initialize Registrar client instance.

        Args:
            netloc: Netloc of registrar service server.
            credentials: Optional credentials.
        """
        super().__init__(
            netloc,
            decoder=RegistrarDecoder(),
            grpc_modules=[fred_api.registry.registrar.service_registrar_grpc_pb2_grpc],
            credentials=credentials,
        )

    async def get_registrar_handles(self) -> List[str]:
        """Get list of all registrar handles."""
        result = await self.call(self.service, 'get_registrar_handles', Empty())
        return cast(List[str], result)

    async def get_registrar_info(self, registrar_handle: str) -> Registrar:
        """Get registrar info by handle."""
        request = RegistrarInfoRequest()
        request.registrar_handle.value = registrar_handle
        result = await self.call(self.service, 'get_registrar_info', request)
        return Registrar(**result)

    async def get_registrar_credit(self, registrar_handle: str) -> Dict[str, Decimal]:
        """Get registrar credit by handle.

        Args:
            registrar_handle: Registrar handle.

        Returns:
            Dictionary with (zone, credit in zone) pairs.
        """
        request = RegistrarCreditRequest()
        request.registrar_handle.value = registrar_handle
        result = await self.call(self.service, 'get_registrar_credit', request)
        return cast(Dict[str, Decimal], result['credit_by_zone'])

    async def get_registrar_zone_access(self, registrar_handle: str) -> Dict[str, bool]:
        """Get registrar zone access.

        Args:
            registrar_handle: Registrar handle.

        Returns:
            Dictionary with (zone, can access zone) pairs.
        """
        request = RegistrarZoneAccessRequest()
        request.registrar_handle.value = registrar_handle
        result = await self.call(self.service, 'get_registrar_zone_access', request)
        return cast(Dict[str, bool], result['access_by_zone'])
