"""Client for SearchKeyset service."""
from datetime import datetime
from typing import Sequence

from fred_api.registry.keyset import service_search_keyset_grpc_pb2_grpc
from fred_api.registry.keyset.service_search_keyset_grpc_pb2 import SearchKeysetHistoryRequest, SearchKeysetRequest

from regal.keyset import KeysetHistoryId, KeysetId

from .common import BaseSearchClient, SearchCurrentResult, SearchDecoder, SearchHistoryResult, SearchResults


class SearchKeysetClient(BaseSearchClient[KeysetId, KeysetHistoryId]):
    """SearchKeyset service gRPC client."""

    grpc_modules = (service_search_keyset_grpc_pb2_grpc,)
    decoder_cls = SearchDecoder[KeysetId, KeysetHistoryId]
    service = 'SearchKeyset'
    limit_name = 'max_number_of_keysets'

    async def search_keyset(
        self,
        query_values: Sequence[str],
        limit: int = None,
        searched_items: Sequence[str] = None
    ) -> SearchResults[SearchCurrentResult[KeysetId]]:
        """Search for keysets based on query.

        Args:
            query_values: List of strings with searched values. Values can be substrings of some keysets fields.
            limit: Maximum number of returned best results.
            searched_items: List of keyset fields that should be included in search.
                If ommited, all the fields will be searched.
        """
        return await self.search_object(
            request=SearchKeysetRequest(), method='search_keyset',
            query_values=query_values, limit=limit, searched_items=searched_items,
        )

    async def search_keyset_history(
        self,
        query_values: Sequence[str],
        limit: int = None,
        searched_items: Sequence[str] = None,
        start: datetime = None,
        end: datetime = None,
    ) -> SearchResults[SearchHistoryResult[KeysetId, KeysetHistoryId]]:
        """Search for keysets history versions based on query.

        Args:
            query_values: List of strings with searched values. Values can be substrings of some keyset fields.
            limit: Maximum number of returned best results.
            searched_items: List of keyset fields that should be included in search.
                If ommited, all the fields will be searched.
            start: Search keyset throughout history starting with this datetime.
                If ommited, there is no lower limit on data age.
            end: Search keyset throughout history ending with this datetime.
                If ommited, search all the way to current data.

        Raises:
            InvalidHistoryInterval: In case of invalid history interval.
        """
        return await self.search_object_history(
            request=SearchKeysetHistoryRequest(), method='search_keyset_history',
            query_values=query_values, limit=limit, searched_items=searched_items, start=start, end=end,
        )
