from datetime import datetime
from unittest import TestCase

import pytz
from fred_api.registry.common_types_pb2 import PlaceAddress
from fred_api.registry.contact.contact_info_types_pb2 import ContactAddress
from fred_api.registry.contact.service_search_contact_grpc_pb2 import SearchContactHistoryReply, SearchContactReply
from fred_api.registry.object_events_pb2 import ObjectEvents as ObjectEventsMessage
from pydantic import BaseModel

from regal.common import Address, BaseId, FuzzyValue, ObjectEvent, ObjectEvents, RegistryDecoder


class BaseIdTest(TestCase):
    def test_validation(self):
        class Model(BaseModel):
            uid: BaseId

        self.assertEqual(Model(uid='007').uid, BaseId('007'))

    def test_repr(self):
        self.assertEqual(repr(BaseId('007')), "BaseId('007')")


class RegistryDecoderTest(TestCase):

    datetime = datetime(2020, 1, 1, tzinfo=pytz.utc)

    def setUp(self):
        self.decoder = RegistryDecoder()
        self.address = Address(
            street=['Floor 42'],
            city='Red Dwarf',
            state_or_province='JMC',
            postal_code='11111',
            country_code='XY',
        )

    def test_decode_fuzzy_value(self):
        self.assertEqual(
            self.decoder.decode(SearchContactReply.Data.FuzzyValue(lower_estimate=10, upper_estimate=100)),
            FuzzyValue(lower_estimate=10, upper_estimate=100)
        )
        self.assertEqual(
            self.decoder.decode(SearchContactHistoryReply.Data.FuzzyValue(lower_estimate=10, upper_estimate=100)),
            FuzzyValue(lower_estimate=10, upper_estimate=100)
        )
        self.assertEqual(
            self.decoder.decode(SearchContactReply.Data.FuzzyValue()),
            FuzzyValue(lower_estimate=0, upper_estimate=0)
        )

    def test_decode_place_address(self):
        address = PlaceAddress()
        address.street[:] = self.address.street
        address.city = self.address.city
        address.state_or_province = self.address.state_or_province
        address.postal_code.value = self.address.postal_code
        address.country_code.value = self.address.country_code

        self.assertEqual(self.decoder.decode(address), self.address)

    def test_decode_contact_address(self):
        self.address.company = 'Jupiter Mining Company'

        address = ContactAddress()
        address.company = self.address.company
        address.street[:] = self.address.street
        address.city = self.address.city
        address.state_or_province = self.address.state_or_province
        address.postal_code.value = self.address.postal_code
        address.country_code.value = self.address.country_code

        self.assertEqual(self.decoder.decode(address), self.address)

    def test_decode_object_event(self):
        self.assertEqual(
            self.decoder.decode(self._make_event('REG-JMC', self.datetime)),
            ObjectEvent(registrar_handle='REG-JMC', timestamp=self.datetime)
        )

    def test_decode_object_events(self):
        events = ObjectEventsMessage()
        events.registered.CopyFrom(self._make_event('REG-JMC', self.datetime))
        events.transferred.CopyFrom(self._make_event('REG-JMC', self.datetime))
        events.updated.CopyFrom(self._make_event('REG-JMC', self.datetime))
        event = ObjectEvent(registrar_handle='REG-JMC', timestamp=self.datetime)

        self.assertEqual(
            self.decoder.decode(events),
            ObjectEvents(
                registered=event,
                transferred=event,
                updated=event,
                unregistered=None,
            )
        )

    @staticmethod
    def _make_event(handle, datetime):
        event = ObjectEventsMessage.EventData()
        event.registrar_handle.value = handle
        event.timestamp.FromDatetime(datetime)
        return event
