from unittest import IsolatedAsyncioTestCase
from unittest.mock import sentinel

from fred_api.registry.contact.contact_history_types_pb2 import (ContactHistoryRecord, ContactHistoryReply,
                                                                 ContactHistoryRequest)
from fred_api.registry.contact.contact_info_types_pb2 import (Birthdate,
                                                              ContactAdditionalIdentifier as
                                                              AdditionalIdentifierMessage, ContactIdReply,
                                                              ContactIdRequest, ContactInfoReply, ContactInfoRequest,
                                                              WarningLetter as WarningLetterMessage)
from fred_api.registry.contact.contact_state_history_types_pb2 import (ContactStateHistory, ContactStateHistoryReply,
                                                                       ContactStateHistoryRequest)
from fred_api.registry.contact.contact_state_types_pb2 import ContactStateReply, ContactStateRequest
from frgal.utils import AsyncTestClientMixin

from regal.contact import (AdditionalIdentifier, Contact, ContactClient, ContactDecoder, ContactHistoryId, ContactId,
                           ContactIdentifierType, WarningLetterPreference)
from regal.tests.common import ObjectClientTestMixin


class ContactDecoderTest(IsolatedAsyncioTestCase):
    """Test ContactDecoder."""

    def setUp(self):
        self.decoder = ContactDecoder()

    def test_additional_identifier(self):
        self.assertEqual(
            self.decoder.decode(AdditionalIdentifierMessage(birthdate=Birthdate(value='2000-01-01'))),
            AdditionalIdentifier(type=ContactIdentifierType.BIRTHDATE, value='2000-01-01')
        )
        self.assertIsNone(self.decoder.decode(AdditionalIdentifierMessage()))

    def test_decode_warning_letter(self):
        self.assertEqual(self.decoder.decode(WarningLetterMessage(preference=0)), WarningLetterPreference.TO_SEND)
        self.assertEqual(self.decoder.decode(WarningLetterMessage(preference=1)), WarningLetterPreference.NOT_TO_SEND)
        self.assertEqual(self.decoder.decode(WarningLetterMessage(preference=2)), WarningLetterPreference.NOT_SPECIFIED)


class TestContactClient(AsyncTestClientMixin, ContactClient):
    """Test ContactClient."""


class ContactClientTest(ObjectClientTestMixin[TestContactClient], IsolatedAsyncioTestCase):
    """Test ContactClient's methods."""

    grpc_namespace = '/Fred.Registry.Api.Contact.Contact'
    get_state_request = ContactStateRequest
    get_state_reply = ContactStateReply
    get_state_method = 'get_contact_state'
    get_history_request = ContactHistoryRequest
    get_history_reply = ContactHistoryReply
    get_history_method = 'get_contact_history'
    get_history_record = ContactHistoryRecord
    get_state_history_request = ContactStateHistoryRequest
    get_state_history_reply = ContactStateHistoryReply
    get_state_history_method = 'get_contact_state_history'
    get_state_history_record = ContactStateHistory.Record

    def setUp(self):
        self.client = TestContactClient(netloc=sentinel.netloc)

    async def test_get_contact_id(self):
        contact_id = ContactId('kryten')
        reply = ContactIdReply()
        reply.data.contact_id.uuid.value = str(contact_id)
        self.client.mock.return_value = reply

        response = await self.client.get_contact_id('HANDLE')
        self.assertEqual(response, contact_id)
        self.assertIsInstance(response, ContactId)

        request = ContactIdRequest()
        request.contact_handle.value = 'HANDLE'
        self.client.mock.assert_called_once_with(
            request,
            method='/Fred.Registry.Api.Contact.Contact/get_contact_id',
            timeout=None,
        )

    async def _test_get_contact_info(self, contact_history_id: ContactHistoryId = None) -> None:
        reply = ContactInfoReply()
        reply.data.contact_id.uuid.value = 'kryten'
        if contact_history_id:
            reply.data.contact_history_id.uuid.value = contact_history_id
        reply.data.contact_handle.value = 'HANDLE'
        reply.data.warning_letter.preference = WarningLetterPreference.NOT_TO_SEND.value
        reply.data.additional_identifier.birthdate.value = '2000-01-01'
        self.client.mock.return_value = reply

        response = await self.client.get_contact_info(ContactId('kryten'), history_id=contact_history_id)

        self.assertEqual(response, Contact(
            contact_id=ContactId('kryten'),
            contact_history_id=contact_history_id,
            contact_handle='HANDLE',
            additional_identifier=AdditionalIdentifier(type=ContactIdentifierType.BIRTHDATE, value='2000-01-01'),
            warning_letter=WarningLetterPreference.NOT_TO_SEND,
        ))

        request = ContactInfoRequest()
        request.contact_id.uuid.value = 'kryten'
        if contact_history_id:
            request.contact_history_id.uuid.value = contact_history_id
        self.client.mock.assert_called_once_with(
            request,
            method='/Fred.Registry.Api.Contact.Contact/get_contact_info',
            timeout=None,
        )

    async def test_get_contact_info(self):
        await self._test_get_contact_info()

    async def test_get_contact_info_with_history_id(self):
        await self._test_get_contact_info(contact_history_id='nova-5')
