from datetime import date, datetime
from typing import Any, Dict, List, Sequence
from unittest import IsolatedAsyncioTestCase
from unittest.mock import sentinel

import pytz
from aioitertools import iter as aiter, list as alist
from fred_api.registry.domain.domain_history_types_pb2 import (DomainHistoryRecord, DomainHistoryReply,
                                                               DomainHistoryRequest)
from fred_api.registry.domain.domain_info_types_pb2 import (DomainIdReply, DomainIdRequest, DomainInfoReply,
                                                            DomainInfoRequest)
from fred_api.registry.domain.domain_state_history_types_pb2 import (DomainStateHistory, DomainStateHistoryReply,
                                                                     DomainStateHistoryRequest)
from fred_api.registry.domain.domain_state_types_pb2 import DomainStateReply, DomainStateRequest
from fred_api.registry.domain.service_admin_grpc_pb2 import (BatchDeleteDomainsReply, BatchDeleteDomainsRequest,
                                                             ManageDomainStateFlagsReply, ManageDomainStateFlagsRequest)
from frgal.utils import AsyncTestClientMixin

from regal.common import DatetimeRange
from regal.contact import ContactId
from regal.domain import Domain, DomainAdminClient, DomainClient, DomainHistoryId, DomainId, NotDeletedReason
from regal.keyset import KeysetId
from regal.nsset import NssetId
from regal.tests.common import ObjectClientTestMixin


class TestDomainClient(AsyncTestClientMixin, DomainClient):
    """Test DomainClient."""


class TestDomainAdminClient(AsyncTestClientMixin, DomainAdminClient):
    """Test DomainAdminClient."""

    exhaust_iterables = True


class DomainClientTest(ObjectClientTestMixin[TestDomainClient], IsolatedAsyncioTestCase):
    """Test DomainClient's methods."""

    grpc_namespace = '/Fred.Registry.Api.Domain.Domain'
    get_state_request = DomainStateRequest
    get_state_reply = DomainStateReply
    get_state_method = 'get_domain_state'
    get_history_request = DomainHistoryRequest
    get_history_reply = DomainHistoryReply
    get_history_method = 'get_domain_history'
    get_history_record = DomainHistoryRecord
    get_state_history_request = DomainStateHistoryRequest
    get_state_history_reply = DomainStateHistoryReply
    get_state_history_method = 'get_domain_state_history'
    get_state_history_record = DomainStateHistory.Record

    def setUp(self):
        self.client = TestDomainClient(netloc=sentinel.netloc)

    async def test_get_domain_id(self):
        domain_id = DomainId('kryten')
        reply = DomainIdReply()
        reply.data.domain_id.uuid.value = str(domain_id)
        self.client.mock.return_value = reply

        response = await self.client.get_domain_id('HANDLE')
        self.assertEqual(response, domain_id)
        self.assertIsInstance(response, DomainId)

        request = DomainIdRequest()
        request.fqdn.value = 'HANDLE'
        self.client.mock.assert_called_once_with(
            request,
            method='/Fred.Registry.Api.Domain.Domain/get_domain_id',
            timeout=None,
        )

    async def _test_get_domain_info(self, domain_history_id: bool = False) -> None:
        reply = DomainInfoReply()
        reply.data.fqdn.value = 'example.com'
        reply.data.domain_id.uuid.value = 'holly'
        reply.data.domain_history_id.uuid.value = 'hilly'
        reply.data.registrant.uuid.value = 'JMC'
        reply.data.nsset.uuid.value = 'red-dwarf'
        reply.data.keyset.uuid.value = 'queeg-500'
        reply.data.sponsoring_registrar.value = 'REG-JMC'
        reply.data.expires_at.FromDatetime(datetime(2030, 1, 1, 10, 0, 4, tzinfo=pytz.utc))
        self.client.mock.return_value = reply

        response = await self.client.get_domain_info(DomainId('holly'),
                                                     history_id=DomainHistoryId('hilly') if domain_history_id else None)

        self.assertEqual(response, Domain(
            fqdn='example.com',
            domain_id=DomainId('holly'),
            domain_history_id=DomainHistoryId('hilly'),
            registrant=ContactId('JMC'),
            nsset=NssetId('red-dwarf'),
            keyset=KeysetId('queeg-500'),
            sponsoring_registrar='REG-JMC',
            expires_at=date(2030, 1, 1),
        ))

        request = DomainInfoRequest()
        request.domain_id.uuid.value = 'holly'
        if domain_history_id:
            request.domain_history_id.uuid.value = 'hilly'
        self.client.mock.assert_called_once_with(
            request,
            method='/Fred.Registry.Api.Domain.Domain/get_domain_info',
            timeout=None,
        )

    async def test_get_domain_info(self):
        await self._test_get_domain_info(False)

    async def test_get_domain_info_with_history_id(self):
        await self._test_get_domain_info(True)

    async def test_get_domain_info_no_expiry_date(self):
        reply = DomainInfoReply()
        reply.data.fqdn.value = 'example.com'
        reply.data.domain_id.uuid.value = 'holly'
        reply.data.domain_history_id.uuid.value = 'hilly'
        reply.data.registrant.uuid.value = 'JMC'
        reply.data.sponsoring_registrar.value = 'REG-JMC'
        self.client.mock.return_value = reply

        response = await self.client.get_domain_info(domain_id='holly')

        self.assertEqual(response, Domain(
            fqdn='example.com',
            domain_id='holly',
            domain_history_id='hilly',
            registrant='JMC',
            sponsoring_registrar='REG-JMC',
            expires_at=None,
        ))


class DomainAdminClientTest(IsolatedAsyncioTestCase):
    """Test DomainAdminClient's methods."""

    def setUp(self):
        self.client = TestDomainAdminClient(netloc=sentinel.netloc)

    async def _test_batch_delete_domains(
            self, domains: Sequence[Domain], reply: BatchDeleteDomainsReply,
            results: List[Dict], requests: List[BatchDeleteDomainsRequest], **kwargs: Any) -> None:
        self.client.mock.return_value = aiter([reply])

        undeleted = [i async for i in self.client.batch_delete_domains(domains, **kwargs)]

        self.assertEqual(undeleted, results)
        self.client.mock.assert_called_once_with(
            tuple(requests),
            method='/Fred.Registry.Api.Domain.Admin/batch_delete_domains',
            timeout=None,
        )

    async def test_batch_delete_domains_empty_result(self):
        domains = [
            Domain(domain_id='holly{}'.format(i), fqdn='example.org', domain_history_id='hilly{}'.format(i))
            for i in range(3)
        ]
        request = BatchDeleteDomainsRequest()
        for domain in domains:
            message = BatchDeleteDomainsRequest.Domain()
            message.domain_id.uuid.value = domain.domain_id
            message.domain_history_id.uuid.value = domain.domain_history_id
            request.domains.append(message)

        await self._test_batch_delete_domains(
            domains=domains,
            reply=BatchDeleteDomainsReply(),
            results=[],
            requests=[request],
        )

    async def test_batch_delete_domains(self):
        reply = BatchDeleteDomainsReply()
        domain = BatchDeleteDomainsReply.NotDeleted()
        domain.domain_id.uuid.value = 'holly'
        domain.reason = NotDeletedReason.DOMAIN_BLOCKED.value
        reply.data.not_deleted_domains.append(domain)
        domains = [
            Domain(domain_id='holly{}'.format(i), fqdn='example.org', domain_history_id='hilly{}'.format(i))
            for i in range(3)
        ]
        request = BatchDeleteDomainsRequest()
        for domain in domains:
            message = BatchDeleteDomainsRequest.Domain()
            message.domain_id.uuid.value = domain.domain_id
            message.domain_history_id.uuid.value = domain.domain_history_id
            request.domains.append(message)

        await self._test_batch_delete_domains(
            domains=domains,
            reply=reply,
            results=[{'domain_id': 'holly', 'reason': NotDeletedReason.DOMAIN_BLOCKED}],
            requests=[request],
        )

    async def test_batch_delete_domains_batch_size(self):
        domains = [
            Domain(domain_id='holly{}'.format(i), fqdn='example.org', domain_history_id='hilly{}'.format(i))
            for i in range(3)
        ]
        requests = []
        for domain in domains:
            request = BatchDeleteDomainsRequest()
            message = BatchDeleteDomainsRequest.Domain()
            message.domain_id.uuid.value = domain.domain_id
            message.domain_history_id.uuid.value = domain.domain_history_id
            request.domains.append(message)
            requests.append(request)

        await self._test_batch_delete_domains(
            domains=domains,
            reply=BatchDeleteDomainsReply(),
            results=[],
            requests=requests,
            batch_size=1,
        )

    async def test_batch_delete_domains_invalid(self):
        # Test call with invalid Domain object - missing domain_history_id.
        domains = [Domain(domain_id='holly', fqdn='example.org', domain_history_id=None)]
        self.client.mock.return_value = aiter([BatchDeleteDomainsReply()])

        with self.assertRaisesRegex(ValueError, "is missing a domain_history_id"):
            await alist(self.client.batch_delete_domains(domains))

        self.client.mock.assert_not_called()

    async def test_manage_domain_state_flags_close(self):
        request_ids = ['request:1', 'request:2']
        self.client.mock.return_value = aiter([ManageDomainStateFlagsReply()])

        response = self.client.manage_domain_state_flags(close_requests=request_ids)

        self.assertEqual(await alist(response), [])
        request = ManageDomainStateFlagsRequest()
        for rid in request_ids:
            flag = request.state_flag_requests_to_close.add()
            flag.uuid.value = rid
        self.client.mock.assert_called_once_with(
            (request,),
            method='/Fred.Registry.Api.Domain.Admin/manage_domain_state_flags',
            timeout=None,
        )

    async def test_manage_domain_state_flags_set(self):
        valid_from = datetime(2022, 1, 1, tzinfo=pytz.utc)
        valid_to = datetime(2022, 2, 1, tzinfo=pytz.utc)
        domains = [
            Domain(domain_id='holly:1', fqdn='holly.example.org'),
            Domain(domain_id='hilly:1', fqdn='hilly.example.org', domain_history_id='hilly:2'),
        ]
        state_flags = {
            'linked': None,
            'serverBlocked': DatetimeRange(start=valid_from, end=valid_to),
        }
        reply = ManageDomainStateFlagsReply()
        for i, domain in enumerate(domains):
            for j, flag in enumerate(state_flags.keys()):
                info = reply.data.state_flag_requests_created.add()
                info.request_id.uuid.value = f'request:{i*10+j}'
                info.domain_id.uuid.value = domain.domain_id
                info.state_flag = flag
        self.client.mock.return_value = aiter([reply])

        response = await alist(self.client.manage_domain_state_flags(domains=domains, state_flags=state_flags))

        self.assertEqual(response, [
            {'domain_id': DomainId('holly:1'), 'request_id': 'request:0', 'state_flag': 'linked'},
            {'domain_id': DomainId('holly:1'), 'request_id': 'request:1', 'state_flag': 'serverBlocked'},
            {'domain_id': DomainId('hilly:1'), 'request_id': 'request:10', 'state_flag': 'linked'},
            {'domain_id': DomainId('hilly:1'), 'request_id': 'request:11', 'state_flag': 'serverBlocked'},
        ])
        self.assertIsInstance(response[0]['domain_id'], DomainId)
        self.assertIsInstance(response[1]['domain_id'], DomainId)

        request = ManageDomainStateFlagsRequest()
        for domain in domains:
            dom = request.state_flag_requests_to_create.domains.add()
            dom.domain_id.uuid.value = domain.domain_id
            if domain.domain_history_id:
                dom.domain_history_id.uuid.value = domain.domain_history_id
        request.state_flag_requests_to_create.state_flags.get_or_create('linked')
        validity = request.state_flag_requests_to_create.state_flags.get_or_create('serverBlocked')
        validity.valid_from.FromDatetime(valid_from)
        validity.valid_to.FromDatetime(valid_to)

        self.client.mock.assert_called_once_with(
            (request,),
            method='/Fred.Registry.Api.Domain.Admin/manage_domain_state_flags',
            timeout=None,
        )

    async def test_manage_domain_state_flags_mismatch(self):
        domains = [Domain(domain_id='holly', fqdn='example.org')]
        state_flags = {'linked': None}

        with self.assertRaises(ValueError):
            await alist(self.client.manage_domain_state_flags(domains=domains))
        with self.assertRaises(ValueError):
            await alist(self.client.manage_domain_state_flags(state_flags=state_flags))
