from decimal import Decimal
from unittest import IsolatedAsyncioTestCase, TestCase
from unittest.mock import sentinel

from fred_api.registry.registrar.registrar_common_types_pb2 import Credit, RegistrarHandle
from fred_api.registry.registrar.registrar_credit_types_pb2 import RegistrarCreditReply, RegistrarCreditRequest
from fred_api.registry.registrar.registrar_info_types_pb2 import RegistrarInfoReply, RegistrarInfoRequest
from fred_api.registry.registrar.registrar_list_types_pb2 import RegistrarListReply
from fred_api.registry.registrar.registrar_zone_access_types_pb2 import (RegistrarZoneAccessReply,
                                                                         RegistrarZoneAccessRequest)
from frgal.utils import AsyncTestClientMixin
from google.protobuf.empty_pb2 import Empty

from regal.common import Address
from regal.registrar import Registrar, RegistrarClient, RegistrarDecoder


class TestRegistrarClient(AsyncTestClientMixin, RegistrarClient):
    """Test RegistrarClient."""


class RegistrarDecoderTest(TestCase):

    def setUp(self):
        self.decoder = RegistrarDecoder()

    def test_decode_credit(self):
        self.assertEqual(self.decoder.decode(Credit(decimal_value='42.19')), Decimal('42.19'))
        self.assertIsNone(self.decoder.decode(Credit()))


class RegistrarClientTest(IsolatedAsyncioTestCase):
    """Test RegistrarClient's methods."""

    def setUp(self):
        self.client = TestRegistrarClient(netloc=sentinel.netloc)

    async def test_get_registrar_handles(self):
        registrar_handles = RegistrarListReply()
        registrar_handles.data.registrar_handles.extend([
            RegistrarHandle(value='REG-SW'),
            RegistrarHandle(value='REG-ST'),
        ])
        self.client.mock.return_value = registrar_handles

        self.assertEqual(await self.client.get_registrar_handles(), ['REG-SW', 'REG-ST'])

        self.client.mock.assert_called_once_with(
            Empty(),
            method='/Fred.Registry.Api.Registrar.Registrar/get_registrar_handles',
            timeout=None,
        )

    async def test_get_registrar_info(self):
        registrar_info = RegistrarInfoReply()
        registrar_info.data.registrar_handle.value = 'REG-SW'
        registrar_info.data.name = 'Star Wars Inc.'
        registrar_info.data.place.street[:] = ['Narrow street 12']
        registrar_info.data.place.city = 'Cloud City'
        registrar_info.data.place.postal_code.value = '12345'
        registrar_info.data.place.country_code.value = 'GE'
        self.client.mock.return_value = registrar_info

        reply = await self.client.get_registrar_info('REG-SW')
        self.assertEqual(reply, Registrar(
            registrar_handle='REG-SW',
            name='Star Wars Inc.',
            place=Address(
                street=['Narrow street 12'],
                city='Cloud City',
                postal_code='12345',
                country_code='GE',
            ),
        ))

        request = RegistrarInfoRequest()
        request.registrar_handle.value = 'REG-SW'
        self.client.mock.assert_called_once_with(
            request,
            method='/Fred.Registry.Api.Registrar.Registrar/get_registrar_info',
            timeout=None,
        )

    async def test_get_registrar_credit(self):
        registrar_credit = RegistrarCreditReply()
        registrar_credit.data.credit_by_zone['cz'].decimal_value = '42000.40'
        registrar_credit.data.credit_by_zone['0.2.4.e164.arpa'].decimal_value = '0.00'
        self.client.mock.return_value = registrar_credit

        self.assertEqual(await self.client.get_registrar_credit('REG-SW'), {
            'cz': Decimal('42000.40'),
            '0.2.4.e164.arpa': Decimal('0.00'),
        })

        request = RegistrarCreditRequest()
        request.registrar_handle.value = 'REG-SW'
        self.client.mock.assert_called_once_with(
            request,
            method='/Fred.Registry.Api.Registrar.Registrar/get_registrar_credit',
            timeout=None,
        )

    async def test_get_registrar_zone_access(self):
        registrar_zone_access = RegistrarZoneAccessReply()
        registrar_zone_access.data.access_by_zone['cz'].has_access_now = True
        registrar_zone_access.data.access_by_zone['0.2.4.e164.arpa'].has_access_now = False
        self.client.mock.return_value = registrar_zone_access

        self.assertEqual(await self.client.get_registrar_zone_access('REG-SW'), {
            'cz': True,
            '0.2.4.e164.arpa': False,
        })

        request = RegistrarZoneAccessRequest()
        request.registrar_handle.value = 'REG-SW'
        self.client.mock.assert_called_once_with(
            request,
            method='/Fred.Registry.Api.Registrar.Registrar/get_registrar_zone_access',
            timeout=None,

        )
